#ifndef STANDARDWRAPPER_H
#define STANDARDWRAPPER_H

#include "Song.h"

#include <future>

/**
  * class StandardWrapper
  *
  * @brief provides common ways of opening, loading and decoding an audio file
  *
  * - automatically normalizes audio during rendering (if provided with an *.ebur128 file)
  * - decides whether to render the whole song to memory at once or not and allocates the pcm buffer (this->data) accordingly
  *
  */


/** @brief this is the common macro responsible for decoding the audio file and correctly filling the PCM buffer
  *
  * usually called within this->render() simply by putting e.g.:
  * 
  *   STANDARDWRAPPER_RENDER(float, 3rd_party_render_function(pcm, framesToDoNow))
  * 
  * "pcm" is the pointer to the buffer where the samples are being stored, as declared below
  * 
  * "framesToDoNow" is the number of frames the render function shall prepare for a given call
  * depending on the semantic of this function you may have to pass the number of items being rendered (i.e. framesToDoNow  * this->Format.Channels)
  * 
  * obviously you may not change the name of these two vars (since they are declared within the scope of the macro), however you may reoder the arguments to fit your render function (this is also the reason why we cant use template function for this purpose, at least I dont see how this should work)
  * 
  * @warning whenever changing this implementation, dont forget LibMadWrapper::render() AND FFMpegWrapper::render(), which does
  * pretty much the same thing, without using this macro
  */
#define STANDARDWRAPPER_RENDER(SAMPLEFORMAT, LIB_SPECIFIC_RENDER_FUNCTION)\
    if(framesToRender==0)\
    {\
        /* render rest of file */ \
        framesToRender = this->getFrames()-this->framesAlreadyRendered;\
    }\
    else\
    {\
        framesToRender = min(framesToRender, this->getFrames()-this->framesAlreadyRendered);\
    }\
    /* audio normalization factor */\
    const float absoluteGain = (numeric_limits<SAMPLEFORMAT>::max()) / (numeric_limits<SAMPLEFORMAT>::max() * this->gainCorrection);\
    const uint32_t Channels = this->Format.Channels(); \
    SAMPLEFORMAT* pcm = static_cast<SAMPLEFORMAT*>(bufferToFill);\
    /* advance the pcm pointer by that many items where we previously ended filling it */\
    pcm += (this->framesAlreadyRendered * Channels) % this->count;\
\
    while(framesToRender>0 && !this->stopFillBuffer)\
    {\
        /* render in chunks of gConfig.FramesToRender size */\
        int framesToDoNow = (framesToRender/gConfig.FramesToRender)>0 ? gConfig.FramesToRender : framesToRender%gConfig.FramesToRender;\
\
        /* call the function whatever is responsible for decoding to raw pcm */\
        LIB_SPECIFIC_RENDER_FUNCTION;\
\
        /* actually do the audio normalization */\
        for(unsigned int i=0; gConfig.useAudioNormalization && i<framesToDoNow*Channels; i++)\
        {\
	    pcm[i] = static_cast<SAMPLEFORMAT>(pcm[i] * absoluteGain);\
        }\
\
        /* advance the pcm pointer by that many items that have just been rendered */\
        pcm += (framesToDoNow * Channels) % this->count;\
        this->framesAlreadyRendered += framesToDoNow;\
\
        framesToRender -= framesToDoNow;\
    }\
 

template<typename SAMPLEFORMAT>
class StandardWrapper : public Song
{
public:

    StandardWrapper(string filename);
    StandardWrapper(string filename, Nullable<size_t> fileOffset, Nullable<size_t> fileLen);

    // forbid copying
    StandardWrapper(StandardWrapper const&) = delete;
    StandardWrapper& operator=(StandardWrapper const&) = delete;

    virtual ~StandardWrapper ();

    void releaseBuffer () noexcept override;

    frame_t getFramesRendered();

    virtual void render(pcm_t* bufferToFill, frame_t framesToRender=0) = 0;

protected:
    // used for double buffering, whenever we were unable to allocate a buffer big enough to hold the whole song in memory
    pcm_t* preRenderBuf = nullptr;

    // a flag that indicates a prematurely abort of async buffer fill
    bool stopFillBuffer = false;

    frame_t framesAlreadyRendered=0;

    template<typename WRAPPERCLASS>
    void fillBuffer(WRAPPERCLASS* context);

    // used for sound normalization
    // usual range: (0,1]
    // can be slightly bigger than 1, depends on the underlying noise normalization algorithm
    // small values indicate a silent source signal resulting in bigger amplification
    // 1.0 means 0 dbFS: no amplification needed in this case ;)
    float gainCorrection = 1.0f;

private:
    future<void> futureFillBuffer;

    void init() noexcept;


};

#endif // STANDARDWRAPPER_H

#include "StandardWrapper_impl.h"
